--[[
   Copyright 2023-2025 Gustaf "Hanicef" Alhäll

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
]]

local tprequests = {}
local tpcooldown = {}
local tpstate = {}
local spawnpoints = {}

local paceframe = {}
local pacetime = {}

local cv_tpcooldowntime = CV_RegisterVar({
	name = "tpcooldowntime",
	defaultvalue = "30",
	flags = CV_NETVAR,
	PossibleValue = CV_Natural,
})

local cv_tptimeout = CV_RegisterVar({
	name = "tptimeout",
	defaultvalue = "18",
	flags = CV_NETVAR,
	PossibleValue = CV_Natural,
})

local cv_visor = CV_RegisterVar({
	name = "visor",
	defaultvalue = "Off",
	flags = 0,
	PossibleValue = CV_OnOff,
})

local cv_minscale = CV_RegisterVar({
	name = "minscale",
	defaultvalue = "0.1",
	flags = CV_NETVAR|CV_FLOAT,
	PossibleValue = CV_Unsigned,
})

local cv_maxscale = CV_RegisterVar({
	name = "maxscale",
	defaultvalue = "20",
	flags = CV_NETVAR|CV_FLOAT,
	PossibleValue = CV_Unsigned,
})

local cv_maxweight = CV_RegisterVar({
	name = "maxweight",
	defaultvalue = "50",
	flags = CV_NETVAR|CV_FLOAT,
	PossibleValue = CV_Unsigned,
})

local cv_maxspeed = CV_RegisterVar({
	name = "maxspeed",
	defaultvalue = 200,
	flags = CV_NETVAR,
	PossibleValue = CV_Unsigned,
})

local cv_chasecam = CV_FindVar("chasecam")

local WALKSPEED = 9 << FRACBITS

if not tofixed then
	-- 2.2.13 compatibility
	rawset(_G, "tofixed", function(n)
		local delim = n:find('.', 0, true)
		if delim then
			local exp = tonumber(n:sub(1, delim-1))
			local frac = tonumber(n:sub(delim+1))
			if exp == nil or frac == nil then
				return nil
			end
			return (exp << FRACBITS) | ((frac << FRACBITS) / (10 ^ (#n-delim)))
		end
		local val = tonumber(n)
		if val == nil then
			return nil
		end
		return val << FRACBITS
	end)
end

local function fixedtostring(v)
	local exp = v >> FRACBITS
	local frac = tostring(((v & (FRACUNIT-1)) * 10000) >> FRACBITS)
	while #frac < 4 do
		frac = "0" .. frac
	end
	frac = frac:gsub("0+$", "")
	if #frac == 0 then
		return exp
	end
	return exp .. "." .. frac
end

local function CheckPaceOverlay(player)
	return player.normalspeed > skins[player.mo.skin].normalspeed or player.jumpfactor > skins[player.mo.skin].jumpfactor
end

local function SpawnPlayerOverlay(player, field, time, condition)
	if player.mo[field] == nil then
		if condition(player) then
			local mo = P_SpawnMobjFromMobj(player.mo, 0, 0, 0, MT_OVERLAY)
			P_SetMobjStateNF(mo, S_INVISIBLE) -- delay update with one tic to avoid triggering state bugs
			mo.destscale = mo.scale
			mo.angle = player.drawangle
			mo.pitch = player.mo.pitch
			mo.roll = player.mo.roll
			mo.spriteroll = player.mo.spriteroll
			mo.skin = player.mo.skin
			mo.color = SKINCOLOR_SUPERSKY1
			mo.colorized = true
			mo.target = player.mo
			mo.frame = (player.mo.frame & FF_FRAMEMASK) | FF_TRANS90
			player.mo[field] = mo
			S_StartSound(player.mo, sfx_monton)
			time[#player] = 0
		end
	end
end

local function HandlePlayerOverlay(player, field, frame, time, skincolor, condition)
	local mo = player.mo[field]
	if mo and mo.valid then
		if frame[#player] == nil then
			frame[#player] = 0
		else
			frame[#player] = frame[#player] + 1
		end
		if frame[#player] < 3*4 then
			mo.color = skincolor + frame[#player] / 3
		elseif frame[#player] > 3*7 then
			mo.color = skincolor
			frame[#player] = 0
		else
			mo.color = skincolor + 4 - (frame[#player] - 3*4) / 3
		end
		P_SetMobjStateNF(mo, player.mo.state)
		mo.destscale = mo.scale
		mo.angle = player.drawangle
		mo.pitch = player.mo.pitch
		mo.roll = player.mo.roll
		mo.spriteroll = player.mo.spriteroll
		mo.sprite = player.mo.sprite
		mo.sprite2 = player.mo.sprite2
		mo.skin = player.mo.skin
		mo.spritexscale = player.mo.spritexscale
		mo.spriteyscale = player.mo.spriteyscale
		if not condition(player) then
			if time[#player] == nil then
				time[#player] = 0
			end
			time[#player] = time[#player] - 1
			if time[#player] <= 0 then
				player.mo[field] = nil -- we don't need to do this, but just in case
				P_RemoveMobj(mo)
				mo = nil
				time[#player] = nil
			end
		else
			if time[#player] == nil then
				time[#player] = 15
			end
			time[#player] = time[#player] + 1
			if time[#player] > 15 then
				time[#player] = 15
			end
		end
		if mo ~= nil then
			mo.frame = (player.mo.frame & FF_FRAMEMASK) | ((9 - time[#player] / 3) << FF_TRANSSHIFT)
		end
	end
end

COM_AddCommand("pace", function (player, speed)
	if not player.mo then
		CONS_Printf(player, "You must be present in a map to change movement pace!")
		return
	end
	if speed == nil then
		if player.normalspeed == skins[player.mo.skin].normalspeed then
			CONS_Printf(player, "You are currently moving at default speed (" .. (player.normalspeed >> FRACBITS) .. " FU/tic)")
		elseif player.normalspeed == skins[player.mo.skin].runspeed - FRACUNIT then
			CONS_Printf(player, "You are currently moving at jogging speed (" .. (player.normalspeed >> FRACBITS) .. " FU/tic)")
		elseif player.normalspeed == WALKSPEED then
			CONS_Printf(player, "You are currently moving at walking speed (10 FU/tic)")
		else
			CONS_Printf(player, "You are currently moving at " .. (player.normalspeed >> FRACBITS) .. " FU/tic")
		end
		return
	end

	if speed:lower() == "walk" then
		player.normalspeed = WALKSPEED
	elseif speed:lower() == "jog" then
		-- subtract one to make sure we don't go into running animation
		player.normalspeed = skins[player.mo.skin].runspeed - FRACUNIT
	elseif speed:lower() == "default" or speed:lower() == "normal" then
		player.normalspeed = skins[player.mo.skin].normalspeed
	elseif speed:lower() == "super" then
		player.normalspeed = skins[player.mo.skin].normalspeed * (player.powers[pw_super] > 0 and 1 or 2)
	else
		local nspeed = tonumber(speed)
		if nspeed ~= nil and nspeed > 0 and nspeed <= cv_maxspeed.value then
			player.normalspeed = nspeed << FRACBITS
		elseif nspeed ~= nil and nspeed > cv_maxspeed.value then
			CONS_Printf(player, "Can't pace to " .. speed .. ": you cannot become faster than " .. cv_maxspeed.value .. "!")
		else
			CONS_Printf(player, speed .. ": not a valid value")
		end
	end

	SpawnPlayerOverlay(player, "paceoverlay", pacetime, CheckPaceOverlay)
end)

COM_AddCommand("jumpheight", function (player, speed)
	if not player.mo then
		CONS_Printf(player, "You must be present in a map to change jump height!")
		return
	end
	if speed == nil then
		CONS_Printf(player, "Your current jump height is " .. (player.jumpfactor >> FRACBITS) .. "." .. (((player.jumpfactor * 10) >> FRACBITS) % 10))
		return
	end
	if speed:lower() == "default" or speed:lower() == "normal" then
		player.jumpfactor = skins[player.mo.skin].jumpfactor
		return
	end
	local nspeed = tofixed(speed)
	if nspeed ~= nil and nspeed > 0 and nspeed <= 5 * FRACUNIT then
		player.jumpfactor = nspeed
	else
		CONS_Printf(player, speed .. ": not a valid value")
	end
end)

COM_AddCommand("knockback", function (player, speed)
	if not player.mo then
		CONS_Printf(player, "You must be present in a map to be knocked back!")
		return
	end
	P_DoPlayerPain(player)
	S_StartSound(player.mo, sfx_shldls)
end)

local fadevalue = 0
local targettime = nil

addHook("MapLoad", function (mapnumber)
	tprequests = {}
	tpcooldown = {}
	targettime = nil
	tpstate = {}
end)

addHook("NetVars", function (network)
	tprequests = network(tprequests)
	tpcooldown = network(tpcooldown)
	targettime = network(targettime)
	spawnpoints = network(spawnpoints)
end)

local function DoTeleport(player, target)
	S_StartSound(player.mo, sfx_s3k75)
	player.mo.target = target.mo
	tpstate[#player] = 1
end

local function FindPlayerByNameOrNumber(name)
	local playernum = tonumber(name)
	if playernum ~= nil and playernum >= 0 and playernum <= 32 then
		local player = players[playernum]
		if not (player and player.valid) then
			return nil
		end
		return  player
	end
	for player in players.iterate do
		if player and player.valid and player.name == name then
			return player
		end
	end
	return nil
end

COM_AddCommand("tpto", function (player, target, force)
	if not player.mo then
		CONS_Printf(player, "You must be present in a map to teleport!")
		return
	end
	if tpcooldown[#player] and leveltime < tpcooldown[#player] then
		CONS_Printf(player, "You must wait " .. ((tpcooldown[#player] - leveltime) / TICRATE) .. " seconds before you can request a teleport again!")
		return
	end
	if target == nil then
		CONS_Printf(player, "tpto <player>: teleports to a player")
		return
	end
	local t = FindPlayerByNameOrNumber(target)
	if not t then
		CONS_Printf(player, "Player " .. target .. " is unknown or invalid")
		return
	end
	target = t
	local playernum = #target

	if not (target and target.valid) then
		CONS_Printf(player, "Player " .. playernum .. " is unknown or invalid")
		return
	end
	if not target.mo then
		CONS_Printf(player, "Player " .. playernum .. " is not present on the map")
		return
	end
	if playernum == #player then
		CONS_Printf(player, "You can't teleport to yourself!")
		return
	end
	if force == "-f" or force == "-force" or target.bot then
		if not IsPlayerAdmin(player) and not target.bot then
			CONS_Printf(player, "Only the server or remote admin can force teleport")
			return
		end
		CONS_Printf(player, "Teleporting to " .. target.name)
		DoTeleport(player, target)
		return
	end
	if tprequests[playernum] then
		CONS_Printf(player, "Someone has already requested a teleport to this player, try again later")
		return
	end
	tprequests[playernum] = { #player, leveltime + cv_tptimeout.value * TICRATE }
	CONS_Printf(player, "Requested teleport to " .. target.name)
	CONS_Printf(target, player.name .. " is requesting to teleport to you")
	S_StartSound(nil, sfx_strpst, target)
end)

COM_AddCommand("tpyes", function (player)
	if not tprequests[#player] then
		CONS_Printf(player, "Nobody has requested a teleport to you")
		return
	end
	if not player.mo then
		CONS_Printf(player, "You cannot accept a teleport when you don't exist in the map!")
		return
	end

	local source = players[tprequests[#player][1]]
	if not (source and source.valid and source.mo) then
		CONS_Printf(player, "Player requesting a teleport no longer exists")
		tprequests[#player] = nil
		return
	end
	CONS_Printf(player, "Accepted teleport from " .. source.name)
	CONS_Printf(source, player.name .. " has accepted your teleport request")
	tprequests[#player] = nil
	DoTeleport(source, player)
end)

COM_AddCommand("tpno", function (player)
	if not tprequests[#player] then
		CONS_Printf(player, "Nobody has requested a teleport to you")
		return
	end

	local source = players[tprequests[#player][1]]
	if not (source and source.valid and source.mo) then
		CONS_Printf(player, "Player requesting a teleport no longer exists")
		tprequests[#player] = nil
		return
	end
	CONS_Printf(player, "Rejected teleport from " .. source.name)
	CONS_Printf(source, player.name .. " has rejected your teleport request")
	tpcooldown[#source] = leveltime + cv_tpcooldowntime.value * TICRATE
	tprequests[#player] = nil
end)

local KEY_ENTER = 13
local KEY_ESCAPE = 27

local KEY_Y = 121
local KEY_N = 110

addHook("KeyDown", function (keyevent)
	if chatactive then
		return
	end
	if not (consoleplayer and consoleplayer.valid) then
		return
	end
	if tprequests[#consoleplayer] then
		if keyevent.num == KEY_Y then
			COM_BufAddText(nil, "tpyes")
			return true
		elseif keyevent.num == KEY_N then
			COM_BufAddText(nil, "tpno")
			return true
		end
	end
end)

addHook("PlayerQuit", function (player)
	-- clear all tp requests so we don't have any lingering teleports
	tprequests[#player] = nil
	for i = 1,#players do
		if tprequests[i] ~= nil and tprequests[i][1] == #player then
			tprequests[i] = nil
		end
	end
end)

addHook("ThinkFrame", function ()
	if consoleplayer and tprequests[#consoleplayer] ~= nil and tprequests[#consoleplayer][2] == leveltime then
		COM_BufAddText(players[#consoleplayer], "tpno")
	end
	if fadevalue >= 10 then
		if not consoleplayer.flashcount then
			-- keep flashing the palette to ensure it doesn't go away
			P_FlashPal(consoleplayer, 6, 1)
		end
	end
end)

hud.add(function (v, stplyr, cam)
	if fadevalue ~= 0 and fadevalue ~= 20 then
		v.fadeScreen(6, min(fadevalue, 20-fadevalue))
	end
	if cv_visor.value then
		if fadevalue == 0 then
			S_StartSound(nil, sfx_GDVISR)
		end
		if fadevalue < 20 then
			fadevalue = fadevalue + 1
		end
	else
		if fadevalue == 20 then
			S_StartSound(nil, sfx_GDVISR)
		end
		if fadevalue > 0 then
			fadevalue = fadevalue - 1
		end
	end
	if targettime and targettime >= leveltime then
		local remain = targettime - leveltime
		local map = V_YELLOWMAP
		if remain <= 5 * TICRATE then
			if remain % TICRATE >= 30 then
				if remain % TICRATE == TICRATE - 1 then
					S_StartSound(nil, sfx_ding)
				end
				map = V_REDMAP
			elseif remain % TICRATE >= 25 then
				map = V_BROWNMAP
			elseif remain % TICRATE >= 20 then
				map = V_ORANGEMAP
			elseif remain % TICRATE >= 15 then
				map = V_PERIDOTMAP
			end
		end
		v.drawString(160, 140, (remain + TICRATE) / TICRATE, map|V_SNAPTOBOTTOM, "center")
	end

	if not tprequests[#consoleplayer] then
		return
	end

	local target = players[tprequests[#consoleplayer][1]]
	if not (target and target.valid) then
		return
	end

	local f = V_YELLOWMAP | V_SNAPTOTOP
	v.drawString(160, 60, target.name .. " want to teleport to you!", f, "small-center")
	v.drawString(160, 65, "Do you wish to accept? (Y/N)", f, "small-center")
end)

COM_AddCommand("shield", function (player, shield)
	if not (player.mo and player.mo.valid) then
		CONS_Printf(player, "You need to be present in the game to use shields!")
		return
	end
	if shield == nil then
		CONS_Printf(player, "Set or remove your shield")
		CONS_Printf(player, "Available shields: none, pity, whirlwind, armageddon, pink, elemental, attract, flameaura, bubblewrap, thundercoin, force")
		return
	end
	shield = shield:lower()
	local sound
	if shield == "none" then
		player.powers[pw_shield] = 0
		P_RemoveShield(player)
		S_StartSound(player.mo, sfx_shldls)
		return
	elseif shield == "pity" then
		player.powers[pw_shield] = SH_PITY
		sound = sfx_shield
	elseif shield == "whirlwind" then
		player.powers[pw_shield] = SH_WHIRLWIND
		sound = sfx_wirlsg
	elseif shield == "armageddon" then
		player.powers[pw_shield] = SH_ARMAGEDDON
		sound = sfx_armasg
	elseif shield == "pink" then
		player.powers[pw_shield] = SH_PINK
		sound = sfx_shield
	elseif shield == "elemental" then
		player.powers[pw_shield] = SH_ELEMENTAL
		sound = sfx_elemsg
	elseif shield == "attract" then
		player.powers[pw_shield] = SH_ATTRACT
		sound = sfx_attrsg
	elseif shield == "flameaura" then
		player.powers[pw_shield] = SH_FLAMEAURA
		sound = sfx_s3k3e
	elseif shield == "bubblewrap" then
		player.powers[pw_shield] = SH_BUBBLEWRAP
		sound = sfx_s3k3f
	elseif shield == "thundercoin" then
		player.powers[pw_shield] = SH_THUNDERCOIN
		sound = sfx_s3k41
	elseif shield == "force" then
		player.powers[pw_shield] = SH_FORCE|1
		sound = sfx_forcsg
	else
		CONS_Printf(player, "Unknown shield " .. shield)
		return
	end
	P_SpawnShieldOrb(player)
	S_StartSound(player.mo, sound)
end)

COM_AddCommand("phase", function (player)
	if not player.mo then
		CONS_Printf(player, "You need to be present in the game to phase!")
		return
	end
	if player.pflags & PF_NOCLIP then
		-- check if it's safe to turn off phasing here
		local subsector = R_PointInSubsectorOrNil(player.mo.x, player.mo.y)
		if R_PointInSubsectorOrNil(player.mo.x, player.mo.y) == nil then
			CONS_Printf(player, "Can't unphase here: you are out of bounds!")
			return
		end
		if player.mo.ceilingz - player.mo.floorz < player.mo.height then
			CONS_Printf(player, "Can't unphase here: not enough headroom to fit!")
			return
		end
	end
	player.pflags = player.pflags ^^ (PF_NOCLIP|PF_GODMODE)
	S_StartSound(player.mo, sfx_GDCLOK)
end)

COM_AddCommand("scale", function (player, scale)
	if not player.mo then
		CONS_Printf(player, "You need to be present in the game to scale!")
		return
	end
	if scale == nil then
		CONS_Printf(player, "scale <number>: set your character's size")
		return
	end
	local n = tofixed(scale)
	if n == nil then
		CONS_Printf(player, scale .. ": not a valid value")
		return
	end
	if n < cv_minscale.value then
		CONS_Printf(player, "Can't scale to " .. scale .. ": you cannot become smaller than " .. fixedtostring(cv_minscale.value) .. "!")
		return
	end
	if n > cv_maxscale.value then
		CONS_Printf(player, "Can't scale to " .. scale .. ": you cannot become larger than " .. fixedtostring(cv_maxscale.value) .. "!")
		return
	end
	-- make sure the player fits the room
	if not (player.pflags & PF_NOCLIP) and player.mo.ceilingz - player.mo.floorz < FixedMul(n, FixedDiv(player.mo.height, player.mo.scale)) then
		CONS_Printf(player, "Can't scale to " .. scale .. ": not enough headroom to fit!")
		return
	end
	player.mo.destscale = n
	CONS_Printf(player, "Scaling to " .. scale)
end)

COM_AddCommand("cloak", function (player, trans)
	if not player.mo then
		CONS_Printf(player, "You need to be present in the game to change your transparency!")
		return
	end
	if trans == nil then
		CONS_Printf(player, "cloak <0-10>: set your character's transparency")
		return
	end
	local n = tonumber(trans)
	if n == nil then
		CONS_Printf(player, trans .. ": not a valid value")
		return
	end
	if n < 0 or n > 10 then
		CONS_Printf(player, "Transparency level must be between 0 and 10!")
		return
	end
	player.ht_trans = n * 6
	S_StartSound(player.mo, sfx_GDCLOK)
end)

COM_AddCommand("super", function (player)
	if not player.mo then
		CONS_Printf(player, "You need to be present in the game to go super!")
		return
	end
	if emeralds ~= 0x7f then
		CONS_Printf(player, "You don't have all the emeralds yet, find them first!")
		return
	end
	if player.powers[pw_super] then
		player.rings = 0
		return
	end
	player.charflags = player.charflags | SF_SUPER
	player.rings = 9999
	P_DoSuperTransformation(player)
end)

rawset(_G, "vcall_countdown_validate", function (player, count)
	if targettime ~= nil and targettime >= leveltime then
		CONS_Printf(player, "A countdown is already ongoing, cancel it with stopcountdown first.")
		return false
	end
	if count == nil then
		CONS_Printf(player, "You didn't specify a time!")
		return false
	end
	local n = tonumber(count)
	if n == nil then
		CONS_Printf(player, count .. " is not a valid value")
		return false
	end
	return true
end)

COM_AddCommand("countdown", function (player, count)
	if targettime ~= nil and targettime >= leveltime then
		CONS_Printf(player, "A countdown is already ongoing, cancel it with stopcountdown first.")
		return
	end
	if count == nil then
		CONS_Printf(player, "You didn't specify a time!")
		return
	end
	local n = tonumber(count)
	if n == nil then
		CONS_Printf(player, count .. " is not a valid value")
		return
	end
	targettime = leveltime + n * TICRATE - 1
end, COM_ADMIN)

rawset(_G, "vcall_stopcountdown_validate", function (player)
	if targettime == nil or targettime < leveltime then
		CONS_Printf(player, "There is no countdown active at the moment.")
		return false
	end
	return true
end)

COM_AddCommand("stopcountdown", function (player)
	if targettime ~= nil and targettime >= leveltime then
		targettime = nil
	else
		CONS_Printf(player, "There is no countdown active at the moment.")
		return
	end
end, COM_ADMIN)

local flipping = {}

COM_AddCommand("flip", function (player)
	if not (player.mo and player.mo.valid) then
		CONS_Printf(player, "You need to be present in the game to flip!")
		return
	end
	player.mo.flags2 = player.mo.flags2 ^^ MF2_OBJECTFLIP
	player.mo.spriteroll = ANGLE_180
	if player.pflags & PF_FLIPCAM or not (cv_chasecam and cv_chasecam.value) then
		player.viewrollangle = ANGLE_180
	end
	S_StartSound(player.mo, sfx_GDFLIP)
	flipping[#player] = true
end)

COM_AddCommand("thrust", function (player, forward, up)
	if not (player.mo and player.mo.valid) then
		CONS_Printf(player, "You need to be present in the game to thrust!")
		return
	end

	if forward == nil then
		CONS_Printf(player, "thrust <forward> [up]: thrusts you forward and optionally up")
		return
	end
	local f = tonumber(forward)
	if f == nil or f < 0 or f > 100 then
		CONS_Printf(player, forward .. ": not a valid value")
		return
	end
	P_Thrust(player.mo, player.mo.angle, f << FRACBITS)
	S_StartSound(player.mo, sfx_GDTRST)
	if up ~= nil then
		local u = tonumber(up)
		if u == nil or u < -100 or u > 100 then
			CONS_Printf(player, up .. ": not a valid value")
			return
		end
		if player.mo.flags2 & MF2_OBJECTFLIP then
			u = -u
		end
		player.mo.momz = player.mo.momz + (u << FRACBITS)
	end
	local mo = P_SpawnMobjFromMobj(player.mo, 0, 0, 0, MT_SCALEGHOST)
	if trans then
		mo.state = S_SCALEGHOST1 + trans - 1
	end
	mo.destscale = mo.scale
	mo.angle = player.drawangle
	mo.pitch = player.mo.pitch
	mo.roll = player.mo.roll
	mo.spriteroll = player.mo.spriteroll
	mo.cusval = player.mo.sprite2
	mo.cvmem = (player.mo.frame & FF_FRAMEMASK)
	mo.skin = player.mo.skin
	mo.color = SKINCOLOR_SUPERPURPLE1
	mo.colorized = true
end)

COM_AddCommand("weight", function (player, value)
	if not (player.mo and player.mo.valid) then
		CONS_Printf(player, "You need to be present in the game to set your weight!")
		return
	end
	if value == nil then
		CONS_Printf(player, "weight <number>: set your own gravity")
		if player.mo.ht_weight == nil then
			CONS_Printf(player, "Your current weight is 1")
		else
			CONS_Printf(player, "Your current weight is " .. player.mo.ht_weight)
		end
		return
	end
	local n = tofixed(value)
	if n == nil or n < 0 then
		CONS_Printf(player, value .. ": not a valid value")
		return
	end
	if n >= cv_maxweight.value then
		CONS_Printf(player, "Can't weigh " .. value .. ": you cannot become heavier than " .. fixedtostring(cv_maxweight.value) .. "!")
		return
	end
	if n == FRACUNIT then
		player.mo.ht_weight = nil
		player.mo.flags = player.mo.flags & MF_NOGRAVITY
		return
	end
	player.mo.ht_weight = n
	player.mo.flags = player.mo.flags | MF_NOGRAVITY
end)

COM_AddCommand("return", function (player, value)
	S_StartSound(player.mo, sfx_s3k75)
	player.mo.targetcoord = spawnpoints[#player]
	tpstate[#player] = 1
end)

COM_AddCommand("getallemeralds", function (player)
	CONS_Printf(player, "You now have all 7 emeralds.")
	emeralds = EMERALD7 * 2 - 1
end, COM_ADMIN)

COM_AddCommand("resetemeralds", function (player)
	CONS_Printf(player, "Emeralds reset to zero.")
	emeralds = 0
end, COM_ADMIN)

local function SetField(player, field, active)
	if active == nil then
		CONS_Printf(player, field .. " <on/off>: enables or disables auto-jump")
		CONS_Printf(player, field .. " is currently " + (player[field] and "enabled" or "disabled"))
		return
	end
	if tonumber(active) or active:lower() == "on" then
		player[field] = true
	elseif tonumber(active) == 0 or active:lower() == "off" then
		player[field] = false
	else
		CONS_Printf(player, field .. " <on/off>: enables or disables auto-jump")
		CONS_Printf(player, field .. " is currently " + (player[field] and "enabled" or "disabled"))
		return
	end
	CONS_Printf(player, field .. " has been " + (player[field] and "enabled" or "disabled"))
end

COM_AddCommand("autojump", function (player, active)
	SetField(player, "autojump", active)
end)

COM_AddCommand("autospin", function (player, active)
	SetField(player, "autospin", active)
end)

-- visual only, don't bother syncing
rawset(_G, "ht_playertrans", { })
local MAXPLAYERTRANS = 35
local playerscaling = { }

for i = 0,#players do
	ht_playertrans[i] = 0
	playerscaling[i] = false
end

addHook("PlayerSpawn", function (player)
	ht_playertrans[#player] = 0
	playerscaling[#player] = false
	tpstate[#player] = nil
	spawnpoints[#player] = nil
	if player.mo then
		player.normalspeed = skins[player.mo.skin].normalspeed
		player.jumpfactor = skins[player.mo.skin].jumpfactor
	else
		player.normalspeed = skins[player.skin].normalspeed
		player.jumpfactor = skins[player.skin].jumpfactor
	end
end)

addHook("PlayerThink", function (player)
	if not player.mo then return end
	if player.playerstate == PST_LIVE and player.mo and (player.mo.eflags & MFE_JUSTHITFLOOR) ~= 0 and not G_EnoughPlayersFinished() then
		if player.autojump and player.cmd.buttons & BT_JUMP then
			player.pflags = player.pflags & ~PF_JUMPDOWN
		elseif player.autospin and player.cmd.buttons & BT_SPIN then
			player.pflags = player.pflags & ~PF_SPINDOWN
		end
	end

	local mobj = player.mo
	if not spawnpoints[#player] then
		spawnpoints[#player] = {x=mobj.x, y=mobj.y, z=mobj.z}
	end
	if (player.pflags & PF_GODMODE) ~= 0 and leveltime % (TICRATE/7) == 0 then
		local sparkle = P_SpawnMobj(mobj.x, mobj.y, mobj.z, MT_IVSP)
		if sparkle and sparkle.valid then
			P_SetScale(sparkle, mobj.scale, true)
		end
	end
	if mobj.ht_weight ~= nil and not P_IsObjectOnGround(mobj) then
		local g = P_GetMobjGravity(mobj)
		if mobj.momz == 0 then
			-- compensate for a special case in the engine
			g = g << 1
		end
		mobj.momz = mobj.momz + FixedMul(g, mobj.ht_weight)
	end
end, MT_PLAYER)

addHook("PostThinkFrame", function ()
	for player in players.iterate do
		if player and player.mo then
			local dest = player.ht_trans
			if not dest then
				dest = 0
			end
			if dest < 30 and player.pflags & PF_NOCLIP then
				dest = 30
			end
			if ht_playertrans[#player] < dest then
				ht_playertrans[#player] = ht_playertrans[#player] + 2
			elseif ht_playertrans[#player] > dest then
				ht_playertrans[#player] = ht_playertrans[#player] - 2
			end
			local trans = ht_playertrans[#player] * 6 / MAXPLAYERTRANS
			player.mo.frame = (player.mo.frame & ~FF_TRANSMASK) | (trans << FF_TRANSSHIFT)
			if player.followmobj and player.followmobj.valid then
				player.followmobj.frame = (player.followmobj.frame & ~FF_TRANSMASK) | (trans << FF_TRANSSHIFT)
			end
			if player.mo.scale ~= player.mo.destscale then
				playerscaling[#player] = true
				if not S_SoundPlaying(player.mo, sfx_s3kc6l) then
					S_StartSound(player.mo, sfx_s3kc6l)
				end
				player.mo.renderflags = player.mo.renderflags | RF_FULLBRIGHT
				player.mo.color = SKINCOLOR_SUPERSILVER1
				player.mo.colorized = true
				if leveltime % 5 == 0 then
					local mo = P_SpawnMobjFromMobj(player.mo, 0, 0, 0, MT_SCALEGHOST)
					if trans then
						mo.state = S_SCALEGHOST1 + trans - 1
					end
					mo.destscale = mo.scale
					mo.angle = player.drawangle
					mo.pitch = player.mo.pitch
					mo.roll = player.mo.roll
					mo.spriteroll = player.mo.spriteroll
					mo.cusval = player.mo.sprite2
					mo.cvmem = (player.mo.frame & FF_FRAMEMASK)
					mo.skin = player.mo.skin
					mo.color = SKINCOLOR_SUPERSILVER1
					mo.colorized = true
				end
			elseif playerscaling[#player] then
				playerscaling[#player] = false
				S_StopSoundByID(player.mo, sfx_s3kc6l)
				player.mo.renderflags = player.mo.renderflags & ~RF_FULLBRIGHT
				player.mo.color = player.skincolor
				player.mo.colorized = false
			end

			HandlePlayerOverlay(player, "paceoverlay", paceframe, pacetime, SKINCOLOR_SUPERSKY1, CheckPaceOverlay)

			if tpstate[#player] == 1 then
				player.mo.spritexscale = player.mo.spritexscale - FRACUNIT / 10
				player.mo.spriteyscale = (FRACUNIT - player.mo.spritexscale) + FRACUNIT
				if player.mo.spritexscale < 0 then
					player.mo.spritexscale = 0
					if player.mo.targetcoord then
						P_SetOrigin(player.mo, player.mo.targetcoord.x, player.mo.targetcoord.y, player.mo.targetcoord.z)
						player.mo.targetcoord = nil
						local mapheader = mapheaderinfo[gamemap]
						S_ChangeMusic(mapheader.musname, true, player, 0, mapheader.muspos)
					elseif player.mo.target and player.mo.target.valid then
						-- only finalize the teleport if the target still exists
						local targetsector = R_PointInSubsectorOrNil(player.mo.target.x, player.mo.target.y)
						if not ((player.pflags & PF_NOCLIP) ~= 0 or targetsector) then
							CONS_Printf(player, "Failed to finalize teleport: target location is not safe to teleport to")
							CONS_Printf(player, "Enable phase first, then try again")
						else
							player.mo.scale = player.mo.target.scale
							player.mo.destscale = player.mo.target.destscale
							player.mo.flags2 = (player.mo.flags2 & ~MF2_OBJECTFLIP) | (player.mo.target.flags2 & MF2_OBJECTFLIP)
							if CV_FindVar("tailspickup").value then
								P_SetOrigin(player.mo, player.mo.target.x, player.mo.target.y, player.mo.target.z)
							else
								-- adjust for player collision
								P_SetOrigin(player.mo, player.mo.target.x, player.mo.target.y, player.mo.target.z + player.mo.target.height)
							end
						end
						player.mo.target = nil
					end
					S_StartSound(player.mo, sfx_s3k75)
					tpstate[#player] = 2
				end
			elseif tpstate[#player] == 2 then
				player.mo.spritexscale = player.mo.spritexscale + FRACUNIT / 10
				player.mo.spriteyscale = (FRACUNIT - player.mo.spritexscale) + FRACUNIT
				if player.mo.spritexscale >= FRACUNIT then
					player.mo.spritexscale = FRACUNIT
					player.mo.spriteyscale = FRACUNIT
					tpstate[#player] = nil
				end
			end

			if flipping[#player] then
				player.viewrollangle = player.viewrollangle / 2
				player.mo.spriteroll = player.mo.spriteroll / 2
				if player.mo.spriteroll == 0 then
					flipping[#player] = nil
				end
			end
		end
	end
end)

function A_CopyPlayerState(mo)
	mo.sprite2 = mo.cusval
	mo.frame = mo.cvmem | (mo.frame & ~FF_FRAMEMASK)
end

if quickmenu then
	quickmenu.add_menu(Submenu("HangoutTools", {
		Entry("Toggle visor", "toggle visor"),
		Submenu("Set velocity", {
			Entry("Walk", "pace walk"),
			Entry("Jog", "pace jog"),
			Entry("Normal", "pace normal"),
			Entry("Super", "pace super"),
		}),
		Entry("Knockback", "knockback"),
		Submenu("Teleport to", function ()
			local menu = {}
			for player in players.iterate() do
				table.insert(menu, Entry(player.name, "tpto \"" .. player.name .. "\""))
			end
			return menu
		end),
		Submenu("Set shield", {
			Entry("None", "shield none"),
			Entry("Pity", "shield pity"),
			Entry("Whirlwind", "shield whirlwind"),
			Entry("Armageddon", "shield armageddon"),
			Entry("Pink", "shield pink"),
			Entry("Elemental", "shield elemental"),
			Entry("Attraction", "shield attract"),
			Entry("Flame aura", "shield flameaura"),
			Entry("Bubblewrap", "shield bubblewrap"),
			Entry("Thundercoin", "shield thundercoin"),
			Entry("Force", "shield force"),
		}),
		Entry("Toggle phase", "phase"),
		Submenu("Scale", {
			Entry("0.1", "scale 0.1"),
			Entry("0.25", "scale 0.25"),
			Entry("0.5", "scale 0.5"),
			Entry("1", "scale 1"),
			Entry("1.5", "scale 1.5"),
			Entry("2", "scale 2"),
			Entry("3", "scale 3"),
			Entry("5", "scale 5"),
			Entry("10", "scale 10"),
			Entry("15", "scale 15"),
			Entry("20", "scale 20"),
		}),
		Submenu("Cloak", function ()
			local menu = {}
			for i = 0,10 do
				table.insert(menu, Entry(tostring(i), "cloak " .. i))
			end
			return menu
		end),
		Entry("Toggle super", "super"),
		Entry("Flip gravity", "flip"),
		Submenu("Thrust", {
			Entry("20 forward, 5 up", "thrust 20 5"),
			Entry("50 forward, 5 up", "thrust 50 5"),
			Entry("100 forward, 5 up", "thrust 100 5"),
			Entry("0 forward, 20 up", "thrust 0 20"),
			Entry("0 forward, 20 down", "thrust 0 -20"),
			Entry("0 forward, 50 up", "thrust 0 50"),
			Entry("0 forward, 50 down", "thrust 0 -50"),
			Entry("Effect only", "thrust 0 0"),
		}),
		Submenu("Set weight", function ()
			local menu = {}
			for i = 0,5 do
				table.insert(menu, Entry(tostring(i), "weight " .. i))
				table.insert(menu, Entry(tostring(i) .. ".25", "weight " .. i .. ".25"))
				table.insert(menu, Entry(tostring(i) .. ".5", "weight " .. i .. ".5"))
				table.insert(menu, Entry(tostring(i) .. ".75", "weight " .. i .. ".75"))
			end
			return menu
		end),
		Entry("Return to spawn", "return"),
		Entry("Give emeralds", "getallemeralds", {admin = true}),
		Entry("Remove emeralds", "resetemeralds", {admin = true}),
		Submenu("Auto-jump", {
			Entry("Enable", "autojump 1"),
			Entry("Disable", "autojump 0"),
		}),
		Submenu("Auto-spin", {
			Entry("Enable", "autospin 1"),
			Entry("Disable", "autospin 0"),
		}),
	}))
end
